#ifndef CONNTRACK_H
#define CONNTRACK_H

#include <asm/types.h>
#include "misc.h"
#include "list.h"
#include "jhash.h"

enum ip_conntrack_dir
{
	IP_CT_DIR_ORIGINAL,
	IP_CT_DIR_REPLY,
	IP_CT_DIR_MAX
};

union ip_conntrack_manip_proto
{
	/* Add other protocols here. */
	u_int16_t all;

	struct {
		u_int16_t port;
	} tcp;
	struct {
		u_int16_t port;
	} udp;
	struct {
		u_int16_t id;
	} icmp;
	struct {
		u_int16_t port;
	} sctp;
};

/* The manipulable part of the tuple. */
struct ip_conntrack_manip
{
	u_int32_t ip;
	union ip_conntrack_manip_proto u;
};

/* This contains the information to distinguish a connection. */
struct ip_conntrack_tuple
{
#ifdef OLDTABLE
	struct list_head list;
#else
	struct ip_conntrack_tuple **hashpointer;
#endif
	struct ip_conntrack_manip src;

	/* These are the parts of the tuple which are fixed. */
	struct {
		u_int32_t ip;
		union {
			/* Add other protocols here. */
			u_int16_t all;

			struct {
				u_int16_t port;
			} tcp;
			struct {
				u_int16_t port;
			} udp;
			struct {
				u_int8_t type, code;
			} icmp;
			struct {
				u_int16_t port;
			} sctp;
		} u;

		/* The protocol. */
		u_int16_t protonum;

		/* Direction */
		enum ip_conntrack_dir dir;
	} dst;
};

struct ip_conntrack
{
	struct ip_conntrack_tuple tuple[IP_CT_DIR_MAX];
};

unsigned int ip_conntrack_hash_rnd;

extern inline unsigned int ip_ct_tuple_equal(struct ip_conntrack_tuple *t1, struct ip_conntrack_tuple *t2)
{
	return t1->src.ip == t2->src.ip
		&& t1->src.u.all == t2->src.u.all
		&& t1->dst.ip == t2->dst.ip
		&& t1->dst.u.all == t2->dst.u.all
		&& t1->dst.protonum == t2->dst.protonum;
}

extern inline uint32_t hash_tuple(struct ip_conntrack_tuple *tuple)
{
	return jhash_3words(tuple->src.ip,
			   (tuple->dst.ip ^ tuple->dst.protonum),
			   (tuple->src.u.all | (tuple->dst.u.all << 16)),
			   ip_conntrack_hash_rnd);
}

#endif
