#define _XOPEN_SOURCE 600
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include "conntrack.h"
#include "newtable.h"

#if 0
#define printd(args...) printf(args)
#else
#define printd(args...)
#endif

void *lookup(struct hashtable_info *info, struct ip_conntrack_tuple *tuple, hash_t hash)
{
	struct hashentry *table = info->hashtable, *bucket;
	unsigned int tablesize = ROOTNUM - 1;
	int hashshift = 0;
	u8 depth = 0;
	unsigned int hashshiftadd = ROOTSHIFT;
	unsigned int bucketnr, i, newbucketnr;
	hashbits_t hashbits;

	printd("\nEntering lookup()\n");
	printd("Looking up entry with hash: %x\n", hash);

restart:
	printd("table: %p\nhashshift: %u\n", table, hashshift);

	bucketnr = (hash >> hashshift) & tablesize;
	bucket = &table[bucketnr];

	hashshift += hashshiftadd;
	if (hashshift > (sizeof(hash_t) * 8 - HASHBITS))
		hashshift -= sizeof(hash_t) * 8 - HASHBITS;

	if (bucket->child) {
		newbucketnr = (hash >> hashshift) & (CHILDNUM - 1);
		prefetchtest(&bucket->child[newbucketnr]);
	}

	hashbits = (hash >> hashshift) & HASHBITMASK;

	printd("looking for hashbits: %x\n", hashbits);

	/* We want to start at entry 1 for depth 0 and entry 0 for all other depths */
	for (i = !depth; i < NUMENTRY; i++) {
		if (!bucket->data.members[i])
			continue;

		printd("checking hashbits of %u\n", i);
		printd("current hashbits: %x\n",
			bucket->hashbits[i] & HASHBITMASK);
		
		if (hashbits == (bucket->hashbits[i] & HASHBITMASK) &&
		    ip_ct_tuple_equal(tuple, bucket->data.members[i])) {
			printd("Confirmed match %p at %u\n",
				bucket->data.members[i], i);
			return container_of(bucket->data.members[i], struct ip_conntrack, tuple[bucket->data.members[i]->dst.dir]);
		} else {
			printd("Miss at %u\n", i);
		}
	}

	if (bucket->child) {
		printd("found child\n");
		table = bucket->child;
		tablesize = CHILDNUM - 1;
		hashshiftadd = CHILDSHIFT;
		depth++;
		goto restart;
	}

	printd("Leaving lookup()\n\n\n");

	return NULL;
}

int alter(struct hashtable_info *info, struct ip_conntrack_tuple *tuple, unsigned long status, hash_t hash)
{
	struct hashentry *table = info->hashtable, *bucket;
	unsigned int tablesize = ROOTNUM - 1;
	int hashshift = 0;
	u8 depth = 0;
	unsigned int hashshiftadd = ROOTSHIFT;
	unsigned int bucketnr, i, newbucketnr;
	hashbits_t hashbits;

	printd("\nEntering alter()\n");
	printd("Looking up entry with hash: %x\n", hash);

restart:
	printd("table: %p\nhashshift: %u\n", table, hashshift);

	bucketnr = (hash >> hashshift) & tablesize;
	bucket = &table[bucketnr];

	hashshift += hashshiftadd;
	if (hashshift > (sizeof(hash_t) * 8 - HASHBITS))
		hashshift -= sizeof(hash_t) * 8 - HASHBITS;

	if (bucket->child) {
		newbucketnr = (hash >> hashshift) & (CHILDNUM - 1);
		prefetchtest(&bucket->child[newbucketnr]);
	}

	hashbits = (hash >> hashshift) & HASHBITMASK;

	printd("looking for hashbits: %x\n", hashbits);

	/* We want to start at entry 1 for depth 0 and entry 0 for all other depths */
	for (i = !depth; i < NUMENTRY; i++) {
		if (!bucket->data.members[i])
			continue;

		printd("checking hashbits of %u\n", i);
		printd("current hashbits: %x\n",
			bucket->hashbits[i] & HASHBITMASK);
		
		if (hashbits == (bucket->hashbits[i] & HASHBITMASK) &&
		    ip_ct_tuple_equal(tuple, bucket->data.members[i])) {
			printd("Confirmed match %p at %u\n",
				bucket->data.members[i], i);
			bucket->hashbits[i] = hashbits | ((status & 1) << HASHBITS);
			return 1;
		} else {
			printd("Miss at %u\n", i);
		}
	}

	if (bucket->child) {
		printd("found child\n");
		table = bucket->child;
		tablesize = CHILDNUM - 1;
		hashshiftadd = CHILDSHIFT;
		depth++;
		goto restart;
	}

	printd("Leaving alter()\n\n\n");

	return 0;
}


int evict(struct hashtable_info *info, struct hashentry *bucket, u8 depth)
{
	unsigned int i, evicted = 0;

	printd("\nEntering evict()\n");
	printd("Evicting from bucket %p at depth %u\n", bucket, depth);

	/* We want to start at entry 1 for depth 0 and entry 0 for all other depths */
	for (i = !depth; i < NUMENTRY; i++) {
		if (!bucket->data.members[i])
			continue;

		/* Try to find an entry to evict */
		if (!(bucket->hashbits[i] & STATUSBITMASK)) {
			/* This is just to make newtest.c happy, should be ripped out */
			bucket->data.members[i]->src.ip = 0;
			bucket->data.members[i]->dst.ip = 0;

			bucket->data.members[i] = NULL;
			info->numentries--;
			evicted++;
		}
	}

	printd("Leaving evict()\n\n\n");

	return evicted;
}

inline unsigned int expand(struct hashtable_info *info, struct hashentry *bucket, uint8_t depth)
{
	struct hashentry *newtable;
	unsigned int ret;

	printd("\nEntering expand()\n");

	if (bucket->child) {
		printf("expand: Error, expanding a bucket with children?\n");
		return 0;
	}

	printd("expand: expanding bucket %p\n", bucket);
	
	//newtable = (void *)malloc(CHILDNUM * sizeof(struct hashentry));
	ret = posix_memalign((void *)&newtable, CACHELINE_REAL, CHILDNUM * sizeof(struct hashentry));
	//if (!newtable)
	if (ret != 0)
		return 0;

	memset(newtable, 0, CHILDNUM * sizeof(struct hashentry));
	
	printd("expand: Newtable: %p\n", newtable);
	
	bucket->child = newtable;
	
	info->maxmem += CHILDNUM * sizeof(struct hashentry);
	info->numchildren++;
	if (info->maxdepth < depth + 1)
		info->maxdepth = depth + 1;

	printd("Leaving expand()\n\n\n");

	return 1;
}

int insert(struct hashtable_info *info, struct ip_conntrack_tuple *tuple, unsigned long status, hash_t hash, unsigned int force_evict)
{
	struct hashentry *table, *bucket, *insertbucket, *lockedbucket = NULL;
	struct hashentry_stack bucketstack[MAXDEPTH];
	unsigned int tablesize;
	int hashshift;
	u8 hashshiftadd = ROOTSHIFT, depth;
	unsigned int bucketnr, insertpos = 0, newbucketnr, evicted = 0;
	int i;
	hashbits_t hashbits, insertbits = 0;

	printd("\nEntering insert()\n");

	printd("Inserting entry %p with hash: %x\n", tuple, hash);
start:
	table = info->hashtable;
	insertbucket = NULL;
	tablesize = ROOTNUM - 1;
	hashshift = 0;
	depth = 0;

restart:
	printd("table: %p\nhashshift: %u\n", table, hashshift);

	bucketnr = (hash >> hashshift) & tablesize;
	bucket = &table[bucketnr];

	bucketstack[depth].bucket = bucket;
	bucketstack[depth].bucketnr = bucketnr;

	if (!lockedbucket) {
		lockedbucket = bucket;
		spin_lock(&(lockedbucket->data.lock));
	}

	hashshift += hashshiftadd;
	if (hashshift > (sizeof(hash_t) * 8 - HASHBITS))
		hashshift -= sizeof(hash_t) * 8 - HASHBITS;

	if (bucket->child) {
		newbucketnr = (hash >> hashshift) & (CHILDNUM -1);
		prefetchtest(&bucket->child[newbucketnr]);
	}

	hashbits = (hash >> hashshift) & HASHBITMASK;
	
	/* We want to start at entry 1 for depth 0 and entry 0 for all other depths */
	for (i = !depth; i < NUMENTRY; i++) {
		if (!bucket->data.members[i]) {
			if (insertbucket == NULL) {
				insertbucket = bucket;
				insertpos = i;
				insertbits = hashbits;
			}
			continue;
		}
		if (hashbits == (bucket->hashbits[i] & HASHBITMASK) &&
		    ip_ct_tuple_equal(tuple, bucket->data.members[i])) {
			printd("Error inserting duplicate entry, "
			       "original found at %u\n", i);
			spin_unlock(&lockedbucket->data.lock);
			return 0;
		}
	}

	/* We have to continue searching for duplicate entries,
	   they might be located somewhere deeper in the trie */
	if (bucket->child) {
		printd("found child\n");
		table = bucket->child;
		tablesize = CHILDNUM - 1;
		hashshiftadd = CHILDSHIFT;
		depth++;
		goto restart;
	}

	if (force_evict) {
		evicted = evict(info, bucket, depth);
		for (i = depth - 1; i >= 0; i--) {
			if (bucketstack[i].bucketnr == info->evictbucketnr || !evicted) {
				evicted += evict(info, bucketstack[i].bucket, i);
				continue;
			}
			break;
		}

		if (!evicted) {
			/* We failed */
			printd("Eviction failed!\n");
			spin_unlock(&lockedbucket->data.lock);
			return 0;
		}
		force_evict = 0;
		goto start;		
	}

	/* Have we found a place to put our new entry? */
	if (insertbucket == NULL) {
		printd("Too many entries in bucket, needs expanding "
		       "(hashshift %u)\n", hashshift);
		if (!expand(info, bucket, depth)) {
			spin_unlock(&lockedbucket->data.lock);
			return 0;
		}

		/* FIXME: no need to restart here, simply calculate the
		   new values one level deeper */
		hashshift -= hashshiftadd;
		if (hashshift < 0)
			hashshift += sizeof(hash_t) * 8 - HASHBITS;
 
		goto restart;
	}

	insertbucket->hashbits[insertpos] = insertbits | ((status & 1) << HASHBITS);
	insertbucket->data.members[insertpos] = tuple;
	tuple->hashpointer = &insertbucket->data.members[insertpos];

	info->numentries++;

	printd("Inserting entry in bucket %p at %u with hashbits %x\n",
		insertbucket, insertpos, insertbits);

	spin_unlock(&lockedbucket->data.lock);

	printd("Leaving insert()\n\n\n");
	return 1;
}

int collapse(struct hashtable_info *info, struct hashentry *parent, u8 parentnrused, struct hashentry *grandparent, hash_t hash)
{
	struct hashentry *table, *bucket, *tmptable, *lparent = parent, *lgrandparent = grandparent;
	int hashshift;
	unsigned int i, tablesize, bucketnr, hashshiftadd, startentry;
	u8 nrused, lparentnrused = parentnrused;

recurse:
	/* this is only true if this is the second pass */
	if (!lparent) {
		hashshift = 0;
		hashshiftadd = ROOTSHIFT;
		tablesize = ROOTNUM - 1;
		startentry = 1;
		table = info->hashtable;
restart:
		nrused = 0;
		bucketnr = (hash >> hashshift) & tablesize;
		bucket = &table[bucketnr];

		for (i = startentry; i < NUMENTRY; i++)
			if (bucket->data.members[i])
				nrused++;

		if (bucket->child) {
			lgrandparent = lparent;
			lparent = bucket;
			lparentnrused = nrused;
			table = bucket->child;
			tablesize = CHILDNUM - 1;
			startentry = 0;
			hashshift += hashshiftadd;
			if (hashshift > (sizeof(hash_t) * 8 - HASHBITS))
				hashshift -= sizeof(hash_t) * 8 - HASHBITS;

			hashshiftadd = CHILDSHIFT;
			goto restart;
		}
	}

	/* Have we reached the top of the tree? */
	if (!lparent)
		return 1;
		
	tmptable = lparent->child;
	lparent->child = NULL;
	/* Job for RCU */
	free(tmptable);

	info->numchildren--;

	/*
	if (lgrandparent) {
		if (lparentnrused == 0)
			lgrandparent->counter--;
		if (lgrandparent->counter == 0) {
			printd("collapse: need recursive collapse\n");
			lgrandparent = lparent = NULL;
			goto recurse;
		}
	}
	*/

	return 1;
}

inline void delete(struct hashtable_info *info, struct ip_conntrack_tuple *tuple)
{
	*tuple->hashpointer = NULL;
	info->numentries--;
}

int register_hashtable(struct hashtable_info *info)
{
	unsigned int i, ret;

	//info->hashtable = (void *)malloc(ROOTNUM * sizeof(struct hashentry));
	ret = posix_memalign((void *)&info->hashtable, CACHELINE_REAL, ROOTNUM * sizeof(struct hashentry));
	if (!info->hashtable) {
	if (ret != 0)
		printf("Error allocating hashtable\n");
		return 0;
	}

	memset(info->hashtable, 0, ROOTNUM * sizeof(struct hashentry));

	info->evictbucketnr = rand() % ROOTNUM;

	for (i = 0; i < ROOTNUM; i++)
		spin_lock_init(&info->hashtable[i].data.lock);

	return 1;
}

int unregister_hashtable(struct hashtable_info *info)
{
	if (info->numentries || info->numchildren) {
		printf("Error unregistering hashtable, not empty. entries: %u children %u\n", info->numentries, info->numchildren);
		return 0;
	}

	free(info->hashtable);

	return 1;
}


