#ifndef NEWTABLE_H
#define NEWTABLE_H

/* ROOTNUM and CHILDNUM _must_ be 2^n */
#define ROOTNUM (1<<ROOTSHIFT)
#define ROOTSHIFT 6
#define CHILDNUM (1<<CHILDSHIFT)
#define CHILDSHIFT 6

#define HASHBITMASK ((1 << HASHBITS) - 1)
#define HASHBITS 7
#define STATUSBITMASK (1 << HASHBITS)

#define HASHALIGN 32

#define MAXDEPTH 7

#define NUMENTRY	((HASHALIGN - sizeof(struct hashentry *)) / (sizeof(hashbits_t) + sizeof(struct ip_conntrack_tuple *)))
#define PADNUM		(HASHALIGN - sizeof(struct hashentry *) - NUMENTRY * (sizeof(hashbits_t) + sizeof(struct ip_conntrack_tuple *)))

typedef u32	hash_t;
typedef u8	hashbits_t;

struct hashtable_info {
	struct hashentry	*hashtable;
	unsigned int		evictbucketnr;

	unsigned int		numchildren;
	unsigned int		numentries;
	unsigned int		maxdepth;
	unsigned int		maxmem;
	unsigned int		insfail;
};

struct hashentry_stack {
	struct hashentry	*bucket;
	u8			bucketnr;
};

struct hashentry {
	unsigned char			filler[PADNUM];
	struct hashentry		*child;
	hashbits_t			hashbits[NUMENTRY];
	/* Assuming that the spinlock gets placed at the start of the union */
	union {
		spinlock_t			lock;
		struct ip_conntrack_tuple	*members[NUMENTRY];
	} data;
} __attribute__((packed));

void *lookup(struct hashtable_info *, struct ip_conntrack_tuple *tuple, hash_t hash);
int alter(struct hashtable_info *, struct ip_conntrack_tuple *tuple, unsigned long status, hash_t hash);
int insert(struct hashtable_info *, struct ip_conntrack_tuple *tuple, unsigned long status, hash_t hash, unsigned int evict);
inline void delete(struct hashtable_info *, struct ip_conntrack_tuple *tuple);

int register_hashtable(struct hashtable_info *);
int unregister_hashtable(struct hashtable_info *);

#endif
