#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include "list.h"
#include "conntrack.h"
#include "oldtable.h"

#if 0
#define printd(args...) printf(args)
#else
#define printd(args...)
#endif

spinlock_t	lock;

inline struct ip_conntrack *__lookup(struct hashtable_info *info, struct ip_conntrack_tuple *tuple, unsigned int bucketnr)
{
	struct ip_conntrack_tuple *c;
	unsigned int longsearch = 0;

	printd("Entering lookup\n");
	printd("lookup: looking for tuple %p in bucket %u\n", tuple, bucketnr);

	list_for_each_entry(c, &(info->hashtable[bucketnr]), list) {
		if (ip_ct_tuple_equal(c, tuple)) {
			if (info->longsearch < longsearch)
				info->longsearch = longsearch;
			return container_of(c, struct ip_conntrack, tuple[c->dst.dir]);
		}
		longsearch++;
	}
	if (info->longsearch < longsearch)
		info->longsearch = longsearch;
	return NULL;
}

struct ip_conntrack *lookup(struct hashtable_info *info, struct ip_conntrack_tuple *tuple, hash_t hash)
{
	unsigned int bucketnr = hash % BUCKETS;
	struct ip_conntrack *ret;

	printd("Entering lookup\n");
	printd("lookup: looking for tuple %p in bucket %u\n", tuple, bucketnr);

	ret = __lookup(info, tuple, bucketnr);

	printd("Leaving lookup\n");

	return ret;
}

int insert(struct hashtable_info *info, struct ip_conntrack_tuple *tuple, hash_t hash)
{
	unsigned int bucketnr = hash % BUCKETS;

	printd("Entering insert\n");
	
	spin_lock(&lock);
	if (__lookup(info, tuple, bucketnr)) {
		spin_unlock(&lock);
		return 0;
	}

	printd("insert: adding entry with tuple %p into bucket %u\n", tuple, bucketnr);

	list_add(&tuple->list, &(info->hashtable[bucketnr]));

	info->numentries++;

	spin_unlock(&lock);
	
	printd("Leaving insert\n");

	return 1;
}

int delete(struct hashtable_info *info, struct ip_conntrack_tuple *tuple)
{
	printd("Entering delete\n");
	printd("delete: removing entry with tuple %p\n", &tuple);

	spin_lock(&lock);
	
	list_del(&tuple->list);

	info->numentries--;

	spin_unlock(&lock);

	printd("Leaving delete\n");

	return 1;
}


int register_hashtable(struct hashtable_info *info)
{
	unsigned int i;

	info->hashtable = (void *)malloc(BUCKETS * sizeof(struct list_head));
	if (!info->hashtable) {
		printf("Error allocating hashtable\n");
		return 0;
	}
	memset(info->hashtable, 0, BUCKETS * sizeof(struct list_head));

	spin_lock_init(&lock);

	for (i = 0; i < BUCKETS; i++)
		INIT_LIST_HEAD(&info->hashtable[i]);

	return 1;
}

int unregister_hashtable(struct hashtable_info *info)
{
	if (info->numentries) {
		printf("Error unregistering hashtable, not empty. entries: %u\n", info->numentries);
		return 0;
	}

	free(info->hashtable);

	return 1;
}
